//////////////////////////////////////////////////////
//			C++ class for complex numbers           //
//////////////////////////////////////////////////////
// written by Alessandro Rosa (zandor_zz@yahoo.it)  //
// code optimized with the help of                  //
// Davide Bucci (davbucci@tiscalinet.it)            //
//////////////////////////////////////////////////////

#include <stdlib.h>
#include <math.h>

#pragma once

#ifndef PI
	#define PI 3.14159265358979323846
#endif

#ifndef E
	#define E  2.71828182845904523536
#endif
/////////////////////////////////////////////////////////
#ifndef DEGREE
	#define DEGREE( rad )  ( fmod( rad , 2.0 * PI ) / PI * 180.0 ) 
#endif

#ifndef RADIANS
	#define RADIANS( deg ) ( fmod( deg , 360.0 ) / 180.0 * PI )
#endif
/////////////////////////////////////////////////////////

class complex
{
public:
	complex();
	~complex();
	complex( const complex& );
	complex( double , double );
	complex( double );

public:
	__inline double radius()	{ return pow( real * real + imag * imag , 0.5 ) ;	}
	__inline double norm()		{ return ( real * real + imag * imag ) ;	        }
	__inline double angle()		{ return atan2( imag, real ) ; }	
	
	__inline friend complex abs( complex z )	{ z.real_abs() ; z.imag_abs() ;   return z; }
	__inline double real_abs()					{ return real = ( real < 0.0 ) ? -real : real ; }
	__inline double imag_abs()					{ return imag = ( imag < 0.0 ) ? -imag : imag ; }

	__inline double GetReal()						{ return real ;	  }
	__inline double GetImag()						{ return imag ;	  }
	__inline double friend GetReal( complex z )		{ return z.real ; } 
	__inline double friend GetImag( complex z )		{ return z.imag ; } 
	
	__inline void plugin( double r , double i )	{ real = r ; imag = i ; }
	__inline complex friend flip( complex z )	{ return complex( z.imag , z.real ) ; }

	__inline friend complex frompolar( double r , double theta )  { return complex( r * cos(theta), r * sin(theta) );  }
    __inline friend complex opposite( complex z )				  { return complex ( -z.real, -z.imag ) ; }
	__inline friend complex conjugate( complex z )			      { return complex( z.real, -z.imag );    }

	__inline friend complex real( complex z )					  {	return complex( z.real , 0.0 ) ;  }
	__inline friend complex imag( complex z )					  {	return complex( z.imag , 0.0 ) ;  }

	__inline bool isGaussianInteger()							  { return ( ( real - double((int)real) ) == 0 && ( imag - double((int)imag) ) == 0 ) ? true : false ; }

public:		// operators
	__inline friend bool operator==( complex a , complex b )	{ return ( ( a.real == b.real ) && ( a.imag == b.imag ) ) ? true : false ; }
	__inline friend bool operator!=( complex a , complex b )	{ return ( a == b ) ? false : true ; }
	__inline friend bool operator<( complex a , complex b )		{ return ( a.norm() < b.norm() ) ? true : false ; }
	__inline friend bool operator>( complex a , complex b )     { return ( a < b ) ? false : true ; }

	void operator+=( complex );
	void operator-=( complex );
	void operator*=( complex );
	void operator/=( complex );

	friend complex operator+( complex , complex );
	friend complex operator+( complex , double  );
	friend complex operator+( double , complex  );

	friend complex operator-( complex );
	
	friend complex operator-( complex , complex );
	friend complex operator-( complex , double  );
	friend complex operator-( double , complex  );

	friend complex operator*( complex , complex );
	friend complex operator*( complex , double  );
	friend complex operator*( double , complex  );

	friend complex operator/( complex , complex );
	friend complex operator/( complex , double  );
	friend complex operator/( double , complex  );

// functions

public:
	friend complex pow( complex , int );
	friend complex pow( complex , double );
	friend complex pow( complex , complex );
	friend complex sqrt( complex );
	friend complex ln( complex );
	friend complex log( complex );
	friend complex exp( complex );

public:	
	friend complex sin( complex );
	friend complex cos( complex );
	friend complex tg( complex );
	friend complex ctg( complex );
	friend complex sec( complex );
	friend complex cosec( complex );

	friend complex sinh( complex );
	friend complex cosh( complex );
	friend complex tgh( complex );
	friend complex ctgh( complex );
	friend complex sech( complex );
	friend complex cosech( complex );

	friend complex asin( complex );
	friend complex acos( complex );
	friend complex actg( complex );
	friend complex atg( complex );
	friend complex asec( complex );
	friend complex acosec( complex );

	friend complex asinh( complex );
	friend complex acosh( complex );
	friend complex actgh( complex );
	friend complex atgh( complex );
	friend complex asech( complex );
	friend complex acosech( complex );

// additionals

public:
	double euclidean_dist( complex );
	double spherical_dist( complex );
    complex rotate( complex, complex, double );
	char* string() ;
	
	friend complex inverse_stereographic( double, double, double ) ;
	friend void stereographic( double, double, double&, double&, double& );

public:
	double real;
	double imag;
};
